"""
Classification related data models
"""
from typing import Dict, List, Optional, Union
from enum import Enum

from pydantic import BaseModel


class ClassificationParadigm(str, Enum):
    """Classification paradigm options"""

    CONTEXT = "context"
    CONTENT = "content"
    INFERRED = "inferred"


class MethodLabel(BaseModel):
    """Classification method output data model"""

    label: Optional[Union[str, List]]
    score: Optional[float]
    aggregated_score: Optional[float]
    position_start: Optional[Union[int, None]] = None
    position_end: Optional[Union[int, None]] = None


class AggregatedMethodLabel(BaseModel):
    """Aggregated classification output data model"""

    label: Optional[str]
    score: Optional[float]
    aggregated_score: Optional[float]
    classification_paradigm: Optional[ClassificationParadigm]

    class Config:
        """Set custom model configuration parameters"""

        use_enum_values = True


class MethodClassification(BaseModel):
    """Classification output data model"""

    input: Union[str, List]
    labels: List


class AggregatedMethodClassification(BaseModel):
    """Aggregated classification output data model"""

    input: Union[str, List]
    labels: List
    aggregation_method: str


class AggregatedSystemClassification(BaseModel):
    """System level classification aggregation"""

    content_input: Optional[List]
    context_input: Optional[List]
    aggregation_method: str = "weighted"
    aggregation_params: Optional[Dict]
    aggregated_labels: List = []



"""
Supports the context classification functionality of fidecls.

Context in this case means metadata about a database such as
schema/table/column name

Analysis of PII through the metadata surrounding a table or
data sample(s)

"""

from typing import Callable, Dict, Iterable, List, Union

from fidescls.cls import classifiers

# pylint: disable=invalid-name

CONTEXT_CLASSIFY_METHODS: Dict[str, Callable] = {
    "similarity": classifiers.similarity_classifier
}


def classify(
    input_data: Union[str, Iterable],
    method_name: str = "similarity",
    method_params: Dict = {},
) -> List:
    """
    Perform context classification using the provided method and associated parameters
    Args:
        input_data: Data to be classified
        method_name: name of the context classification method to use (default: similarity)
        method_params: parameters to pass along to the classification method

    Returns:
        The classification response
    """
    if method_name not in CONTEXT_CLASSIFY_METHODS.keys():
        raise ValueError(f"{method_name} is not a valid context classification method")
    return CONTEXT_CLASSIFY_METHODS[method_name](input_data, **method_params)



"""
This file holds the decision functionality.  What to do after data goes through the classifier
"""

import logging
from enum import Enum
from typing import cast, Dict, List, Union

from fidescls.cls import entity_map, models as _cls_models


class DecisionTypeEnum(Enum):
    """
    Enum object to capture valid decision types
    """

    PASS_THROUGH = "pass-through"
    DIRECT_MAPPING = "direct-mapping"


def rank_classification_results(
    cls_results: Union[
        List,
        Dict[str, List[_cls_models.MethodClassification]],
    ],
    desc: bool = True,
    sort_attribute: str = "score",
) -> Union[
    List[_cls_models.MethodClassification],
    Dict[str, List[_cls_models.MethodClassification]],
]:
    """
    Order the classification results by classification score
    Args:
        cls_results: The result output from classification
        desc: flag to set ranking order
        sort_attribute: attribute by which to rank

    Returns:
        the same as the classification output but with the results ranked in order
    """
    if isinstance(cls_results, dict):
        for _, cls_output in cls_results.items():
            for classification in cls_output:
                classification.labels.sort(
                    key=lambda x: getattr(x, sort_attribute), reverse=desc
                )
    else:
        for result in cls_results:
            result.labels.sort(key=lambda x: getattr(x, sort_attribute), reverse=desc)
    return cls_results


def decide(
    cls_result: Union[
        List[_cls_models.MethodClassification],
        Dict[str, List[_cls_models.MethodClassification]],
    ],
    decision_method: str,
) -> Union[
    List[_cls_models.MethodClassification],
    Dict[str, List[_cls_models.MethodClassification]],
]:
    """
    Determine which decision method to use
    Args:
        cls_result: the classification result from the method used
        decision_method: the choice in decision method to use

    Returns:

    """
    # Get decision method by string
    if decision_method == DecisionTypeEnum.DIRECT_MAPPING.value:
        return direct_mapping(cls_result)
    if decision_method == DecisionTypeEnum.PASS_THROUGH.value:
        return pass_through(cls_result)
    raise ValueError(
        f"Invalid Decision Type. Must be one of {[i.value for i in DecisionTypeEnum]}"
    )


def pass_through(
    cls_result: Union[
        List[_cls_models.MethodClassification],
        Dict[str, List[_cls_models.MethodClassification]],
    ]
) -> Union[
    List[_cls_models.MethodClassification],
    Dict[str, List[_cls_models.MethodClassification]],
]:
    """
    Raw pass-through of the result from the classifier
    Args:
        cls_result: the result output from a classifier

    Returns:

    """
    return cls_result


def convert_entity_to_category(
    category: str,
    method_output: Union[_cls_models.MethodLabel, _cls_models.AggregatedMethodLabel],
) -> Union[_cls_models.MethodLabel, _cls_models.AggregatedMethodLabel]:
    """
    convert a classification method output label to a data category while retaining the output attributes
    Args:
        category: the data category to use as the new label
        method_output: the classification method output with a PII entity label

    Returns:
        the converted classification method output with the data category as its label
    """
    if isinstance(method_output, _cls_models.AggregatedMethodLabel):
        return _cls_models.AggregatedMethodLabel(
            label=category,
            score=method_output.score,
            classification_paradigm=method_output.classification_paradigm,
        )
    return _cls_models.MethodLabel(
        label=category,
        score=method_output.score,
        position_start=method_output.position_start,
        position_end=method_output.position_end,
    )


def convert_entities_to_categories(
    classify_output: Union[
        _cls_models.MethodClassification, _cls_models.AggregatedMethodClassification
    ],
    entity_category_mapping: Union[Dict, None] = None,
) -> List:
    """
    Convert the PII entity labels from a classification result to a mapped alternative
    Args:
        classify_output: the classification result
        entity_category_mapping: the mapping to use to convert the pii entity labels

    Returns:
        List of the newly converted classification method labels
    """
    if not entity_category_mapping:
        entity_category_mapping = entity_map.ENTITIES

    mapped_results = []
    for classification_label in classify_output.labels:
        try:
            if not classification_label.label:
                mapped_results.extend([classification_label])
            else:
                mapped_labels = cast(
                    List, entity_category_mapping[cast(str, classification_label.label)]
                )
                mapped_results.extend(
                    [
                        convert_entity_to_category(label, classification_label)
                        for label in mapped_labels
                    ]
                )
        except KeyError:
            mapped_results = []
    return mapped_results


def direct_mapping(
    classification_output: Union[List, Dict[str, List]],
    entity_category_mapping: Union[Dict, None] = None,
) -> Union[List, Dict[str, List]]:
    """
    make decision based on direct mapping to default fides taxonomy data_categories

    Args:
        classification_output: the result output from a classifier
        entity_category_mapping: dictionary mapping the class labels to data_categories

    Returns:
        classification result with mapped labels
    """
    logging.debug(
        f"Performing direct-mapping on classification results - result type {type(classification_output)}"
    )
    if isinstance(classification_output, dict):
        for classification_results in classification_output.values():
            logging.debug(f"Mapping data categories for {classification_results}")
            for classification_result in classification_results:
                classification_result.labels = convert_entities_to_categories(
                    classification_result,
                    entity_category_mapping=entity_category_mapping,
                )
    else:
        for classification_result in classification_output:
            logging.debug(f"Mapping data categories for {classification_result}")
            classification_result.labels = convert_entities_to_categories(
                classification_result, entity_category_mapping=entity_category_mapping
            )
    return classification_output



"""
Contains the groups and setup for the CLI.
"""

import click

from fidescls._version import get_versions
from fidescls.cli.classify import classify
from fidescls.cli.scan import db
from fidescls.cli.api import api
from fidescls.cli import logger as _cli_logger

logger = _cli_logger.config_logger()


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(name="fidescls", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def cli(ctx: click.Context) -> None:
    """
    The Fidescls command line interface to interact
    with the PII classification functionality
    """
    ctx.ensure_object(dict)


@cli.command(name="version", help="get the version of fidescls")
def version() -> None:
    """
    get the version of fidescls
    """
    click.secho(f'{get_versions()["version"]}')


# Add new commands to the cli contex
cli.add_command(classify)
cli.add_command(db)
cli.add_command(api)



"""
Utility functionality to support the scanning of a database for PII
"""

from collections import defaultdict
from typing import Callable, Dict, List, Union
import warnings

import click
import pandas as pd
from sqlalchemy.engine.base import Engine
from tqdm import tqdm

from fidescls.cls import content, context, aggregation, models as _cls_models
from fidescls.utils import databases as _db

try:
    from fideslang import default_taxonomy
except ModuleNotFoundError:
    raise click.ClickException("Unable to find fideslang, please install!")


def nested_default_dict() -> defaultdict[defaultdict]:  # type: ignore
    """
    Create an infinitely nestable dictionary
    """
    return defaultdict(nested_default_dict)


def db_context_classify(
    metadata_df: pd.DataFrame,
    context_method: Callable = context.classify,
    method_params: Dict = {},
    hide_progress: bool = False,
) -> defaultdict:
    """
    Perform context classification given a database's schema metadata
    Args:
        metadata_df: dataframe containing database metadata
        context_method: method to use when performing context classification
        method_params: context classification method kwargs
        hide_progress: flag to disable displaying of progress bars

    Returns:
        Nested dictionary containing schema:table:column = [classification results]
    """
    if not hide_progress:
        click.secho("Classifying database context...", fg="white")

    if not method_params.get("possible_targets"):
        taxonomy = default_taxonomy.DEFAULT_TAXONOMY
        method_params["possible_targets"] = [
            i.fides_key for i in taxonomy.data_category
        ]

    context_labels = nested_default_dict()
    for _, row in tqdm(
        metadata_df.iterrows(),
        total=metadata_df.shape[0],
        disable=hide_progress,
        desc="Tables",
        position=0,
    ):
        for column in tqdm(
            row["columns"],
            disable=hide_progress,
            desc="Columns",
            leave=False,
            position=1,
        ):
            context_labels[row["schema"]][row["table"]][column] = list(
                context_method(
                    f"{row['schema']}.{row['table']}.{column}",
                    method_name="similarity",
                    method_params=method_params,
                )
            )
    return context_labels


def db_content_classify(
    db_engine: Engine,
    metadata_df: pd.DataFrame,
    num_samples: int = 1,
    content_method: Callable = content.classify,
    method_params: dict = {},
    hide_progress: bool = False,
) -> defaultdict[defaultdict]:  # type: ignore
    """
    Perform content classification on a sample dataset taken from tables in a database
    Args:
        db_engine: sqlalchemy database engine
        metadata_df: dataframe containing database metadata
        num_samples: number of samples to take from each table
        content_method: method to use when performing content classification
        method_params: content classification method kwargs
        hide_progress: flag to disable displaying of progress bars

    Returns:
        Nested dictionary containing schema:table:column = [classification results]
    """
    if not hide_progress:
        click.secho(
            f"Classifying database content using {num_samples} sample(s)..", fg="white"
        )

    data_samples_cls = nested_default_dict()
    for _, row in tqdm(
        metadata_df.iterrows(),
        total=metadata_df.shape[0],
        disable=hide_progress,
        desc="Samples",
        position=0,
    ):
        table_samples = _db.get_table_samples(
            db_engine,
            schema_name=row["schema"],
            table_name=row["table"],
            num_samples=num_samples,
        )
        if table_samples:
            data_samples_cls[row["schema"]][row["table"]] = content_method(
                table_samples, **method_params
            )
        else:
            # no samples were returned from the query to the table
            data_samples_cls[row["schema"]][row["table"]] = {
                k: [_cls_models.MethodClassification(input="", labels=[])]
                for k in row["columns"]
            }
    return data_samples_cls


def db_aggregate_system(
    scan_output: Dict,
    aggregation_method: str,
    aggregation_params: Dict,
    hide_progress: bool = False,
    structure_source: str = "context",
) -> defaultdict:
    """
    Perform system level aggregation on the database scanning and classification output
    Args:
        scan_output: The classification output from a database scan
        aggregation_method: The method by which to aggregate (default: weighted)
        aggregation_params: The kwargs to be passed onto the aggregation method
        hide_progress: Flag to high processing progress bars
        structure_source: classification paradigm name to use when traversing db structure

    Returns:
        The aggregated database classification
    """

    if not hide_progress:
        click.secho(
            "Aggregating context and content classification output..", fg="white"
        )
    aggregated_output = nested_default_dict()
    for schema_name, schema in tqdm(
        scan_output[structure_source].items(),
        total=len(scan_output["context"]),
        disable=hide_progress,
        desc="Schema",
        position=0,
    ):
        for table_name, table in tqdm(
            schema.items(),
            total=len(schema),
            disable=hide_progress,
            desc="Tables",
            position=1,
            leave=False,
        ):
            for column_name, _ in tqdm(
                table.items(),
                total=len(table),
                disable=hide_progress,
                desc="Columns",
                position=2,
                leave=False,
            ):
                aggregated_output[schema_name][table_name][
                    column_name
                ] = aggregation.aggregate_system(
                    scan_output["context"][schema_name][table_name][column_name],
                    check_column_continuity(
                        scan_output, schema_name, table_name, column_name
                    ),
                    aggregation_method=aggregation_method,
                    aggregation_params=aggregation_params,
                )
    return aggregated_output


def check_column_continuity(
    scan_output: Dict,
    schema_name: str,
    table_name: str,
    column_name: str,
    paradigm_to_check: str = "content",
) -> List:
    """
    Ensure that a column is present in both the content and context database scanning output.
    If not, add said column with an empty classification output.
    Args:
        scan_output: The output from a database scanning operation
        schema_name: The schema to check for parity
        table_name: The table to check for parity
        column_name: The column to check for parity
        paradigm_to_check: The paradigm (content, context) on which to check parity

    Returns:
        The list of classifications for a given column with any missing columns
        given an empty classification using the structure_source as the ground truth
    """
    try:
        return scan_output[paradigm_to_check][schema_name][table_name][column_name]
    except KeyError:
        warnings.warn(
            f"Mismatch in column parity for {'.'.join([schema_name, table_name, column_name])}! "
            f"Inferring no {paradigm_to_check} classification"
        )
        return [
            _cls_models.AggregatedMethodClassification(
                input="",
                labels=[
                    _cls_models.AggregatedMethodLabel(
                        classification_paradigm="inferred"
                    )
                ],
                aggregation_method="inferred",
            )
        ]


def label_database(
    connection_string: str,
    classify_context: bool = False,
    classify_content: bool = False,
    aggregate_output: bool = False,
    context_method: Callable = context.classify,
    content_method: Callable = content.classify,
    aggregation_method: Union[None, str] = "weighted",
    context_method_params: Dict = {},
    content_method_params: Dict = {
        "decision_method": "direct-mapping",
        "aggregation_method": "mean",
    },
    aggregation_params: Dict = {"top_n": 3},
    num_samples: int = 10,
    hide_progress: bool = False,
) -> Union[Dict[str, Dict], defaultdict]:
    """
    Perform context and/or content classification on a mysql or postgres database
    Args:
        connection_string:  sqlalchemy compatible database connection string
        classify_context: flag to perform context classification (default: False)
        classify_content: flag to perform content classification (default: False)
        aggregate_output: flag to perform system level classification aggregation (default: False)
        context_method: method to use when performing context classification
        content_method: kwargs to pass to the context classification method
        aggregation_method: method used to perform system aggregation
        context_method_params: method kwargs to pass to the context classification method
        content_method_params: method kwargs to pass to the context classification method
        aggregation_params: method kwargs to pass to the system aggregation method
        num_samples: number of samples to use when collecting data for content classification
        hide_progress: flag to disable progress bars and feedback

    Returns:
        Dictionary containing the results for content and/or context classification
        for each column in each table of the database
    """
    # check to see that any inspection is happening
    if not classify_context and not classify_content:
        raise click.exceptions.ClickException("No classification method selected!")

    db_engine = _db.get_engine(connection_string, verbose=True)
    db_metadata = _db.scan_database_metadata(
        connection_string, hide_progress=hide_progress
    )
    metadata_df = _db.metadata_to_dataframe(db_metadata)

    dataset_labels = {}
    if classify_context and context_method:
        context_cls_results = db_context_classify(
            metadata_df,
            context_method=context_method,
            method_params=context_method_params,
            hide_progress=hide_progress,
        )
        dataset_labels["context"] = context_cls_results

    if classify_content and content_method:
        content_cls_results = db_content_classify(
            db_engine,
            metadata_df,
            num_samples=num_samples,
            content_method=content_method,
            method_params=content_method_params,
            hide_progress=hide_progress,
        )
        dataset_labels["content"] = content_cls_results
    if aggregate_output and aggregation_method:
        if not {"context", "content"}.issubset(dataset_labels.keys()):
            raise ValueError(
                "System aggregation only available when both content and context classification are specified!"
            )
        return db_aggregate_system(
            dataset_labels,
            aggregation_method=aggregation_method,
            aggregation_params=aggregation_params,
            hide_progress=hide_progress,
        )
    return dataset_labels  # type: ignore



"""
Base text classification functionality
"""
import logging
from typing import cast, Iterable, Union, List, Dict

from presidio_analyzer import AnalyzerEngine, RecognizerResult
from presidio_analyzer.nlp_engine import NlpEngineProvider

from fidescls.cls import config as _cconf, models as _cls_models
from fidescls.utils import fides_nlp as _nlp

logger = logging.getLogger(__name__)


class BatchAnalyzerEngine(AnalyzerEngine):
    """
    Class inheriting from AnalyzerEngine and adds the functionality
    to analyze lists or dictionaries.
    https://microsoft.github.io/presidio/samples/python/batch_processing/#set-up-classes-for-batch-processing
    """

    def __init__(self, small_model: bool = True):
        """
        Use the small english model instead of the default large one for speed
        TODO: make this configurable
        """
        configuration = {
            "nlp_engine_name": "spacy",
            "models": [
                {
                    "lang_code": "en",
                    "model_name": "en_core_web_sm" if small_model else "en_core_web_lg",
                }
            ],
        }
        provider = NlpEngineProvider(nlp_configuration=configuration)
        super().__init__(nlp_engine=provider.create_engine())

    def analyze_list(
        self,
        list_of_texts: List[str],
        language: str = _cconf.LANGUAGE,
        **kwargs: List[str],
    ) -> List[List[RecognizerResult]]:
        """
        Analyze a list of strings

        Args:
            list_of_texts: A list containing strings to be analyzed
            language: language string
            **kwargs: Additional parameters for the `AnalyzerEngine.analyze` method

        Returns:

        """

        list_results = []
        for text in list_of_texts:
            results = (
                self.analyze(text, language, **kwargs) if isinstance(text, str) else []
            )
            list_results.append(results)
        return list_results

    def analyze_dict(
        self,
        input_dict: Dict[str, Union[object, List[object]]],
        language: str = _cconf.LANGUAGE,
        **kwargs: List[str],
    ) -> List[Dict[str, Union[str, List, RecognizerResult]]]:
        """
        Analyze a dictionary of keys (strings) and values (either object or List[object]).
        Non-string values are returned as is.

        Args:
            input_dict: The input dictionary for analysis
            language: language string
            **kwargs: Additional keyword arguments for the `AnalyzerEngine.analyze` method

        Returns:

        """
        results = []
        for key, values in input_dict.items():
            item_result: Union[
                List[None], List[RecognizerResult], List[List[RecognizerResult]]
            ]
            if not values:
                item_result = []
            else:
                if isinstance(values, str):
                    item_result = self.analyze(values, language, **kwargs)
                elif isinstance(values, List):
                    item_result = self.analyze_list(
                        list_of_texts=cast(List[str], values),
                        language=language,
                        **kwargs,
                    )
                else:
                    item_result: List[None] = []  # type: ignore
            results.append(
                {"key": key, "value": values, "recognizer_results": item_result}
            )
        return results


class BaseTextClassifier(BatchAnalyzerEngine):
    """
    Presidio based text PII classifier/analyzer
    """

    def classify(
        self,
        raw_text: Union[str, List[str], Dict[str, List]],
        language: str = _cconf.LANGUAGE,
        **kwargs: List[str],
    ) -> Union[
        List[Union[List[RecognizerResult], Dict[str, List[RecognizerResult]]]],
        List[RecognizerResult],
    ]:
        """
        High level, standardized function call for text classification
        Args:
            raw_text: input data to be classified
            language: language string
            **kwargs: keyword arguments that get passed onto the `AnalyzerEngine.analyze` method

        Returns:

        """
        classify_result: Union[
            List[Union[List[RecognizerResult], Dict[str, List[RecognizerResult]]]],
            List[RecognizerResult],
        ]
        logger.debug(
            f"BaseTextClassifier classify inputs: {raw_text}, {language}, {kwargs}"
        )
        if isinstance(raw_text, list):
            logger.debug("Classifying a list")
            classify_result = self.analyze_list(raw_text, language, **kwargs)
        elif isinstance(raw_text, dict):
            logger.debug("Classifying a dict")
            raw_dict = cast(Dict[str, Union[object, List[object]]], raw_text)
            classify_result = self.analyze_dict(raw_dict, language, **kwargs)
        else:
            logger.debug("Classifying a string")
            classify_result = self.analyze(raw_text, language, **kwargs)

        logger.debug(f"Classify result: {classify_result}")
        return classify_result


def similarity_classifier(
    input_text: Union[str, Iterable],
    possible_targets: Iterable = (),
    top_n: Union[int, None] = 1,
    remove_stop_words: bool = False,
    pii_threshold: Union[float, None] = _cconf.PII_DEFAULT_THRESHOLD,
) -> List[_cls_models.MethodClassification]:
    """
    Perform similarity classification between an argument and a set of possible targets
    Args:
        input_text: Text for which to check similarity
        possible_targets: list of candidates to compare against
        top_n: number of similar responses to return in descending
        order. If None, all results will be returned
        remove_stop_words: flag to remove stop words from the text
        pii_threshold: similarity is classified as 'not-pii' if score is below this threshold. 'None' = ignored

    Returns:
        List of MethodClassifications containing the text and top_n most similar possible_targets
    """
    logging.debug("Performing context similarity classification...")

    if not isinstance(input_text, list):
        input_text = [input_text]

    target_docs = {
        possible_target: _nlp.tokenize(
            _nlp.preprocess_text(possible_target, remove_stop_words=remove_stop_words)
        )
        for possible_target in possible_targets
    }

    input_docs = {
        text: _nlp.tokenize(
            _nlp.preprocess_text(text, remove_stop_words=remove_stop_words)
        )
        for text in input_text
    }

    similarities = {}
    for raw_input, input_doc in input_docs.items():
        target_similarities = {}
        for raw_target, target_doc in target_docs.items():
            similarity_score = round(_nlp.similarity(input_doc, target_doc), 4)
            if pii_threshold and similarity_score < pii_threshold:
                logging.debug(
                    f"Similarity score [{similarity_score}] between "
                    f"[{raw_input}] and [{raw_target}] below threshold [{pii_threshold}], skipping"
                )
                continue
            target_similarities[raw_target] = similarity_score
        # grab top n similar entries
        similarities[raw_input] = _nlp.get_top_n_dict(target_similarities, top_n=top_n)

    # format the output to meet the model standard
    formatted_output = []
    for sim_word, similar in similarities.items():
        formatted_output.append(
            _cls_models.MethodClassification(
                input=sim_word,
                labels=[
                    _cls_models.MethodLabel(
                        label=label_score[0],
                        score=label_score[1],
                        position_start=None,
                        position_end=None,
                    )
                    for label_score in similar
                ],
            )
        )

    return formatted_output



"""
API related data models
"""
from typing import Dict, List, Iterable, Optional, Union
from pydantic import BaseModel
from fidescls.api import config as _api_conf


class SimilarityParams(BaseModel):
    """Data model for the similarity method params"""

    possible_targets: Iterable
    top_n: Optional[Union[int, None]] = 1
    remove_stop_words: bool = False
    pii_threshold: Optional[Union[float, None]] = _api_conf.PII_DEFAULT_THRESHOLD


class ContentParams(BaseModel):
    """Data model for the content paradigm options"""

    language: str = "en"
    model: Optional[Union[str, None]] = None
    decision_method: str = "direct-mapping"
    aggregation_method: str = "mean"
    infer_not_pii: bool = True


class AggregationPayload(BaseModel):
    """Data model for classification aggregation"""

    method: str
    method_params: Dict


class ContentPayload(BaseModel):
    """Data model for the content paradigm"""

    data: Union[List, Dict]
    method: str = "default"
    method_params: ContentParams
    aggregation: Optional[AggregationPayload]


class ContextPayload(BaseModel):
    """Data model for the context paradigm"""

    data: Union[str, List]
    method: str = "similarity"
    method_params: SimilarityParams
    aggregation: Optional[AggregationPayload]


class ClassifyPayload(BaseModel):
    """Data model for classification request payloads"""

    context: Optional[ContextPayload]
    content: Optional[ContentPayload]


class RequestPayload(BaseModel):
    """Input Data Payload"""

    data: List[ClassifyPayload]
    data_aggregation: Optional[AggregationPayload]



"""
Natural Language Processing functionality specific to fidescls
"""

import logging
import re
import warnings
from typing import Dict, Iterable, List, Mapping, Tuple, Union
import spacy
from spacy import tokenizer, symbols, tokens


NLP_MODEL = "en_core_web_lg"
NLP = spacy.load(NLP_MODEL)
# capture a list of special tokens that we'd want to make sure get handled as expected
SPECIAL_TOKENS = {"id": ("id",)}


def get_nlp_tokenizer(
    special_cases: Union[None, Mapping[str, Iterable[str]]] = None,
    vocab: spacy.vocab.Vocab = NLP.vocab,
) -> tokenizer.Tokenizer:
    """
    Setup custom tokenizer with special cases
    https://spacy.io/usage/linguistic-features/#special-cases
    TODO: Find a way to do this once instead of for each tokenization

    Args:
        special_cases: array of strings that need special handling to make sure they get tokenized properly
        e.g.
                {
                'gimme': (gim, me),
                'id': (id)
                }
        vocab: spacy vocab language
    """
    if not special_cases:
        special_cases = SPECIAL_TOKENS
    vocab_tokenizer = tokenizer.Tokenizer(vocab)
    for word, special_tokens in special_cases.items():
        special_case = [{symbols.ORTH: token} for token in special_tokens]
        logging.debug(f"Adding special case for tokenization: {special_case}")
        vocab_tokenizer.add_special_case(word, special_case)
    return vocab_tokenizer


def tokenize(
    text: str, special_cases: Union[None, Mapping[str, Iterable[str]]] = None
) -> spacy.tokens.doc.Doc:
    """
    Convert a string to a spacy tokenized document
    Args:
        text: the text string to be tokenized
        special_cases: special token cases used to initialize the nlp framework

    Returns:
        the input text tokens
    """
    # TODO: inspect latency with instantiating the tokenizer with every call
    vocab_tokenizer = get_nlp_tokenizer(
        special_cases=special_cases if special_cases else SPECIAL_TOKENS
    )
    return vocab_tokenizer(text)


def get_top_n_dict(
    dict_to_sort: Dict, top_n: Union[int, None] = 1, descending: bool = True
) -> List[Tuple[str, float]]:
    """
    Sort and filter dictionary by value
    Args:
        dict_to_sort: dictionary to sort by its value
        top_n: number of top results to filter down to (set to None for all values returned in sorted order)
        descending: which direction to sort the dictionary values

    Returns:
        A dictionary with the `top_n` entries, sorted by value in `descending` order
    """
    return sorted(dict_to_sort.items(), key=lambda x: x[1], reverse=descending)[:top_n]


def similarity(
    text1: Union[str, tokens.doc.Doc], text2: Union[str, tokens.doc.Doc]
) -> float:
    """
    Perform a spacy similarity (cosine) between two strings or docs
    Args:
        text1: a string or spacy doc to compare
        text2: a string or spacy doc to compare

    Returns:
        Similarity value [0,1] - 0: not similar, 1: exact same
    """
    logging.debug(f"Checking similarity between [{text1}] and [{text2}]")
    # TODO figure out how to best handle empty vectors (word not in vocabulary)
    with warnings.catch_warnings():
        warnings.filterwarnings("error")
        try:
            if isinstance(text1, str):
                text1 = tokenize(text1)
            if isinstance(text2, str):
                text2 = tokenize(text2)
            return text1.similarity(text2)
        except UserWarning as warn:
            logging.debug(
                f"Similarity is 0 due to either [{text1}] or [{text2}] not in vocab! {warn}"
            )
            return 0.0


def preprocess_text(
    text: str,
    remove_stop_words: bool = True,
    removal_pattern: str = r"_|\/|\.(?!\s|$)|\d",
    replacement_pattern: str = " ",
) -> str:
    """
    Pre-process text by removing and splitting based on a removal pattern
    Args:
        text: text to be processed
        remove_stop_words: flag to remove spacy default stop words
        removal_pattern: regex pattern of which characters to remove
        replacement_pattern:

    Returns:

    """
    # TODO: the text gets tokenized here and has to be re-tokenized after, efficiency to be gained here
    doc = tokenize(re.sub(removal_pattern, replacement_pattern, text).lower())
    if remove_stop_words:
        # TODO: research around if we may want to be a bit more particular about stop words
        return " ".join(
            [token.text for token in doc if token.text not in NLP.Defaults.stop_words]
        )
    return " ".join([token.text for token in doc])



"""
CLI command wrappers for spinning up the Fidecsl API
"""
import subprocess as sp
import click


@click.group(name="api")
def api() -> None:
    """
    A CLI command group focused on database scanning for classification
    """
    # pylint: disable=invalid-name
    ...


@api.command(name="start")
@click.pass_context
@click.option(
    "-h", "--host", "host", type=str, default="0.0.0.0", help="host to run the api on"
)
@click.option(
    "-p", "--port", "port", type=int, default=8765, help="port number the api will use"
)
def start(ctx: click.Context, host: str, port: int) -> None:
    """
    Start the fidescls api server
    """
    sp.check_call(
        f"uvicorn --host {host} --port {port} --reload fidescls.api.main:app",
        shell=True,
    )



"""
Configure the python logger for the fidescls cli
"""

import logging
from logging.handlers import RotatingFileHandler
from fidescls.cli import config as _conf


def config_logger(logging_level: str = "DEBUG") -> logging.Logger:
    """
    Create and configure a logger that creates a rotating log file
    Args:
        logging_level: string for the logging level to be used

    Returns:
        python logger with a rotating file handler
    """
    logger = logging.getLogger(_conf.LOGGER_NAME)
    logger.setLevel(logging_level)

    handler = RotatingFileHandler(_conf.LOG_FILENAME, maxBytes=int(50e6), backupCount=1)
    handler.setLevel(logging_level)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.handlers = []
    logger.addHandler(handler)
    return logger



"""
High level functionality around performing PII classification
"""
import logging
from typing import cast, List, Dict, Union
import numpy as np

from presidio_analyzer import RecognizerResult

from fidescls.cls import (
    decision,
    classifiers,
    config as _cls_config,
    models as _cls_models,
    aggregation as _cls_agg,
    entity_map as _cls_entities,
)

logger = logging.getLogger(__name__)


def format_method_labels(
    method_result: Union[None, RecognizerResult]
) -> _cls_models.MethodLabel:
    """
    Standardized the format of the recognizer results
    Args:
        method_result: A recognizer result
    """
    if method_result:
        return _cls_models.MethodLabel(
            label=method_result.entity_type,
            score=np.round(method_result.score, decimals=4),
            position_start=method_result.start,
            position_end=method_result.end,
        )
    return _cls_models.MethodLabel()


def format_cls_output(
    input_data: Union[List[str], Dict[str, List]],
    cls_output: List[Union[List[RecognizerResult], Dict[str, List[RecognizerResult]]]],
) -> Union[
    List[_cls_models.MethodClassification],
    Dict[str, List[_cls_models.MethodClassification]],
]:
    """
    Format the content classifier's output to meet the spec
    Args:
        input_data: The data that was classified
        cls_output: The output of the classifier

    Returns:
        Formatted classification output to conform to the spec
    """
    formatted_output: Union[
        List[_cls_models.MethodClassification],
        Dict[str, List[_cls_models.MethodClassification]],
    ]
    labels: Union[None, _cls_models.MethodLabel, List[_cls_models.MethodLabel]]

    if isinstance(cls_output, list) and isinstance(cls_output[0], list):
        formatted_output = []
        for i, input_value in enumerate(input_data):
            labels = [format_method_labels(cls_out) for cls_out in cls_output[i]]
            formatted_output.append(
                _cls_models.MethodClassification(input=input_value, labels=labels)
            )
    else:
        formatted_output = {}
        for cls_out in cls_output:
            cls_out_casted = cast(
                Dict[str, Union[str, List[RecognizerResult]]], cls_out
            )
            cls_out_result = cast(
                List[RecognizerResult], cls_out_casted["recognizer_results"]
            )
            cls_out_key = cast(str, cls_out_casted["key"])
            parsed_output = []
            for i, input_value in enumerate(cls_out_casted["value"]):
                if cls_out_result[i]:
                    labels = [
                        format_method_labels(j)
                        if cls_out_result[i]
                        else _cls_models.MethodLabel()
                        for j in cls_out_result[i]
                    ]
                else:
                    labels = []
                parsed_output.append(
                    _cls_models.MethodClassification(input=input_value, labels=labels)
                )
            formatted_output[cls_out_key] = parsed_output
    return formatted_output


def classify(
    data: Union[str, List[str], Dict[str, List]],
    language: str = _cls_config.LANGUAGE,
    model: str = None,
    decision_method: str = "direct-mapping",
    aggregation_method: Union[str, None] = None,
    infer_not_pii: bool = True,
) -> Union[
    List[_cls_models.MethodClassification],
    Dict[str, List[_cls_models.MethodClassification]],
]:
    """
    Perform a PII classification on data using the specified model and decision method
    Args:
        data: Data to be classified (list or dictionary)
        language: language for text model purposes. supported: ['en']
        model: specify which model to use. Default (None, "") is the BaseTextClassifier
        decision_method: specify which decision method to use. supported:
            'pass-through': direct pass-through of results from the model
            'direct-mapping': the mapped value described by an entity map
        aggregation_method: the method to use when aggregating the classification output
        infer_not_pii: flag to infer not pii score based on occurrence if a not-pii label is not a class label
    Returns:
        The output from the classifier in the format dependent on the chosen `decision_method`

    """
    if isinstance(data, str):
        data = [data]
    logger.debug(f"Classifying {data} using decision: {decision_method}")
    if not model:
        cls = classifiers.BaseTextClassifier()
    else:
        raise NotImplementedError("Only default model supported currently!")

    # perform the classification
    cls_label = cls.classify(
        data, language=language, entities=list(_cls_entities.ENTITIES.keys())
    )

    # format the classification output
    cls_label = format_cls_output(data, cls_label)

    # aggregate the classification output
    if aggregation_method:
        cls_label = _cls_agg.aggregate_classification(
            cls_label,
            aggregation_method=aggregation_method,
            infer_not_pii=infer_not_pii,
        )
    # implement any decision based on return if requested
    decision_result = decision.decide(cls_label, decision_method)
    return decision.rank_classification_results(decision_result)



"""
Common options and arguments for cli commands
"""

from typing import Callable

import click

from fidescls.cls import config as _cls_config


def decision_method(command: Callable) -> Callable:
    """
    determine which decision method to use
    """
    return click.option(
        "-d",
        "--decision-method",
        "decision_method",
        type=str,
        default="direct-mapping",
        help="Decision method [pass-through | direct-mapping]",
    )(command)


# language option
def language(command: Callable) -> Callable:
    """
    Add an option for language selection
    """
    return click.option(
        "-l",
        "--language",
        "language",
        type=str,
        default=_cls_config.LANGUAGE,
        help="Text language for NLP processing",
    )(command)


def model(command: Callable) -> Callable:
    """
    Add an option for ml model selection
    """
    return click.option(
        "-m", "--model", "model", default=None, type=str, help="ML model to use"
    )(command)


def top_n(command: Callable) -> Callable:
    """
    Add an option to select the number of classification results to return
    """
    return click.option(
        "-n",
        "--top-n",
        "top_n",
        type=click.IntRange(1, clamp=True),
        default=1,
        help="""Number of results to return from the PII classification process""",
    )(command)


def num_samples(command: Callable) -> Callable:
    """
    Add an option to select the number of classification results to return
    """
    return click.option(
        "-s",
        "--samples",
        "num_samples",
        type=click.IntRange(1, clamp=True),
        default=10,
        help="""Number of samples to use for content classification""",
    )(command)


def possible_targets(command: Callable) -> Callable:
    """
    Add an option to provide an array of similarity targets to compare against
    """
    return click.option(
        "-p",
        "--possible-targets",
        "possible_targets",
        type=str,
        nargs=1,
        default="",
        help='Comma separated string containing a list of possible targets (i.e. "target1, target2")',
    )(command)


def remove_stop_words(command: Callable) -> Callable:
    """
    Add an option to provide an array of similarity targets to compare against
    """
    return click.option(
        "-s",
        "--remove-stop-words",
        "remove_stop_words",
        default=False,
        is_flag=True,
        help="Flag to remove stop words when processing",
    )(command)


def pii_threshold(command: Callable) -> Callable:
    """
    Provide a minimum threshold score for context PII classification. None = no threshold
    """
    return click.option(
        "-t",
        "--pii-threshold",
        "pii_threshold",
        default=None,
        help="Minimum threshold for context pii classification. No threshold by default",
    )(command)


def use_content(command: Callable) -> Callable:
    """
    Add option flag to use content classification
    """
    return click.option(
        "--content",
        "use_content",
        default=False,
        is_flag=True,
        help="Flag to use content classification",
    )(command)


def use_context(command: Callable) -> Callable:
    """
    Add option flag to use context classification
    """
    return click.option(
        "--context",
        "use_context",
        default=False,
        is_flag=True,
        help="Flag to use context classification",
    )(command)


def use_aggregation(command: Callable) -> Callable:
    """
    Add option flag to use classification output aggregation
    """
    return click.option(
        "-a",
        "--aggregate",
        "use_aggregation",
        is_flag=True,
        default=False,
        help="Flag to use classification output aggregation",
    )(command)


def aggregation_method(command: Callable) -> Callable:
    """
    Method of to use for classification output aggregation
    """
    return click.option(
        "-aa",
        "--aggregation-method",
        "aggregation_method",
        type=str,
        nargs=1,
        default="weighted",
        help="Aggregation method to use for content and context classification output aggregation (default: weighted)",
    )(command)



"""
Contains the code that sets up the API.
"""
import os
from typing import Dict

from fastapi import FastAPI, status
from uvicorn import Server, Config

import fidescls
from fidescls.api import config as _acon
from fidescls.api import classify, logger as _api_logger


LOGGING_LEVEL = os.environ.get("LOGGING_LEVEL", "DEBUG").upper()
logger = _api_logger.config_logger(logging_level=LOGGING_LEVEL)

app = FastAPI(title="fidescls")


def configure_routes() -> None:
    """Include all routers not defined here."""
    for router in classify.routers:
        app.include_router(router)


configure_routes()


@app.get(
    "/health",
    response_model=Dict[str, str],
    responses={
        status.HTTP_200_OK: {
            "content": {
                "application/json": {
                    "example": {
                        "status": "healthy",
                        "version": "1.0.0",
                    }
                }
            }
        }
    },
    tags=["Health"],
)
async def health() -> Dict:
    "Confirm that the API is running and healthy."
    return {
        "status": "healthy",
        "version": fidescls.__version__,
    }


def start_webserver() -> None:
    """Run the webserver."""
    server = Server(Config(app, host="0.0.0.0", port=_acon.PORT))
    server.run()



"""
Classifier API Endpoint
"""
import logging
from typing import cast, List, Dict, Union

from fastapi import APIRouter

from fidescls.api import config as _aconf, models as _api_models
from fidescls.cls import (
    models as _cls_models,
    aggregation as _cls_agg,
    context,
    content,
)

logger = logging.getLogger(_aconf.LOGGER_NAME)

# pylint: disable=invalid-name

RESPONSE = Union[
    Dict[
        str,
        Union[
            List,
            Dict[str, List],
        ],
    ],
    _cls_models.AggregatedSystemClassification,
]
RESPONSE_LIST = List[RESPONSE]

routers = []
for resource_type in _aconf.RESOURCES:
    router = APIRouter(tags=["Classify", resource_type], prefix=f"/{resource_type}")

    @router.post(f"/{_aconf.CLASSIFY_ENDPOINT}")
    async def classify(
        payload: _api_models.RequestPayload,
    ) -> RESPONSE_LIST:
        """
        Perform PII classification
        Args:
            payload: request payload containing fields described in the ClassifyPayload data model

        Returns:
            The result of the classification

        Raises:
            HTTPException of status code
        """
        responses: RESPONSE_LIST = []

        if payload.data:
            for input_data in payload.data:
                response: RESPONSE = {}

                if input_data.context:
                    if input_data.context.method == "similarity":
                        logger.debug(
                            f"Similarity Request Payload:\n{input_data.context.dict()}"
                        )
                        context_labels = context.classify(
                            input_data.context.data,
                            method_name="similarity",
                            method_params={
                                "possible_targets": input_data.context.method_params.possible_targets,
                                "top_n": input_data.context.method_params.top_n,
                                "remove_stop_words": input_data.context.method_params.remove_stop_words,
                                "pii_threshold": input_data.context.method_params.pii_threshold,
                            },
                        )
                        response["context"] = context_labels  # type: ignore
                if input_data.content:
                    if input_data.content.method == "default":
                        logger.debug(
                            f"Content Request Payload:\n{input_data.content.dict()}"
                        )
                        content_labels = content.classify(
                            input_data.content.data,
                            language=input_data.content.method_params.language,
                            model=input_data.content.method_params.model,
                            decision_method=input_data.content.method_params.decision_method,
                            aggregation_method=input_data.content.method_params.aggregation_method,
                            infer_not_pii=input_data.content.method_params.infer_not_pii,
                        )
                        response["content"] = content_labels  # type: ignore
                if payload.data_aggregation:
                    response = _cls_agg.aggregate_system(
                        context_classification=cast(List, response["context"]),  # type: ignore
                        content_classification=cast(List, response["content"]),  # type: ignore
                        aggregation_method=payload.data_aggregation.method,
                        aggregation_params=payload.data_aggregation.method_params,
                    )
                responses.append(response)
        else:
            raise ValueError("No input data provided for classification!")
        return responses

    routers.append(router)



"""
CLI command wrappers for classify functionality
"""

import pprint
from typing import cast
import click

from fidescls.cli import options as _opt

from fidescls.cls import models as _cls_models


@click.group(name="classify")
def classify() -> None:
    """
    A CLI command group focused on PII classification methods
    """
    ...


@classify.command(name="content")
@click.pass_context
@_opt.language
@_opt.model
@_opt.decision_method
@click.argument("data", type=str)
def content(
    ctx: click.Context, data: str, language: str, model: str, decision_method: str
) -> None:
    """
    Detect and classify PII content

    DATA: data string to be processed (comma separated)
    """
    from fidescls.cls import content as _content

    cls_output = _content.classify(
        [i.strip() for i in data.split(",")],
        language=language,
        model=model,
        decision_method=decision_method,
    )
    click.secho(
        pprint.pformat(
            [cast(_cls_models.MethodClassification, i).dict() for i in cls_output]
        ),
        fg="white",
    )


@classify.command(name="context")
@click.pass_context
@_opt.top_n
@_opt.possible_targets
@_opt.remove_stop_words
@_opt.pii_threshold
@click.argument("data", type=str)
def context(
    ctx: click.Context,
    data: str,
    top_n: int,
    possible_targets: str,
    remove_stop_words: bool,
    pii_threshold: float,
) -> None:
    """
    PII context classification

    DATA: data string to be processed (comma separated)
    """
    from fidescls.cls import context as _context

    if not possible_targets:
        raise click.BadArgumentUsage(
            "No possible targets were provided! They are required for context classification."
        )
    cls_output = _context.classify(
        [i.strip() for i in data.split(",")],
        method_name="similarity",
        method_params={
            possible_targets: [i.strip() for i in possible_targets.split(",")],
            top_n: top_n,
            remove_stop_words: remove_stop_words,
            pii_threshold: pii_threshold,
        },
    )
    click.secho(pprint.pformat([i.dict() for i in cls_output]), fg="white")



import pathlib
from setuptools import setup, find_packages

import versioneer

here = pathlib.Path(__file__).parent.resolve()
long_description = open("docs/fidescls/docs/installation/pypi.md").read()

# Requirements
install_requires = open("requirements.txt").read().strip().split("\n")
dev_requires = open("dev-requirements.txt").read().strip().split("\n")

setup(
    name="fidescls",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="CLI for Fidescls",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ethyca/fidescls",
    entry_points={"console_scripts": ["fidescls=fidescls.cli.context:cli"]},
    python_requires=">=3.8, <4",
    package_dir={"fidescls": "fidescls"},
    packages=find_packages(),
    author="Ethyca, Inc.",
    author_email="fidesteam@ethyca.com",
    license="Apache License 2.0",
    install_requires=install_requires,
    dev_requires=dev_requires,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
    ],
)



"""
Basic Mapping of persidio entities and fides default taxonomy

TODO: convert this to yaml file so customized mappings can be appended

"""

ENTITIES = {
    "CREDIT_CARD": [
        "account.payment.financial_account_number",
        "user.provided.identifiable.financial.account_number",
    ],
    "CRYPTO": [
        "account.payment.financial_account_number",
        "user.provided.identifiable.financial.account_number",
    ],
    "DATE_TIME": ["user.provided.identifiable.date_of_birth"],
    "EMAIL_ADDRESS": [
        "account.contact.email",
        "user.provided.identifiable.contact.email",
    ],
    "IBAN_CODE": [
        "account.payment.financial_account_number",
        "user.provided.identifiable.financial.account_number",
    ],
    "IP_ADDRESS": ["user.derived.identifiable.device.ip_address"],
    "NRP": [
        "user.provided.identifiable.political_opinion",
        "user.provided.identifiable.religious_belief",
        "user.derived.identifiable.race",
    ],
    "LOCATION": [
        "account.contact.street",
        "account.contact.city",
        "account.contact.state",
        "account.contact.country",
        "account.contact.postal_code",
        "user.provided.identifiable.contact.street",
        "user.provided.identifiable.contact.city",
        "user.provided.identifiable.contact.state",
        "user.provided.identifiable.contact.country",
        "user.provided.identifiable.contact.postal_code"
        "user.derived.identifiable.location",
    ],
    "PERSON": ["user.provided.identifiable.name"],
    "PHONE_NUMBER": [
        "account.contact.phone_number",
        "user.provided.identifiable.contact.phone_number",
    ],
    "MEDICAL_LICENSE": [
        "user.provided.identifiable.health_and_medical",
        "user.provided.identifiable.government_id",
    ],
    "US_BANK_NUMBER": [
        "account.payment.financial_account_number",
        "user.provided.identifiable.financial.account_number",
    ],
    "US_DRIVER_LICENSE": [
        "user.provided.identifiable.government_id.drivers_license_number"
    ],
    "US_ITIN": [
        "user.provided.identifiable.government_id.national_identification_number"
    ],
    "US_PASSPORT": ["user.provided.identifiable.government_id.passport_number"],
    "US_SSN": [
        "user.provided.identifiable.government_id.national_identification_number"
    ],
    "UK_NHS": [
        "user.provided.identifiable.government_id.national_identification_number",
        "user.provided.identifiable.health_and_medical",
    ],
    "AU_ABN": [
        "user.provided.identifiable.government_id.national_identification_number"
    ],
    "AU_ACN": [
        "user.provided.identifiable.government_id.national_identification_number"
    ],
    "AU_TFN": [
        "user.provided.identifiable.government_id.national_identification_number"
    ],
    "AU_MEDICARE": [
        "user.provided.identifiable.government_id.national_identification_number",
        "user.provided.identifiable.health_and_medical",
    ],
}



"""
Classification output aggregation functionality
"""

# pylint: disable=invalid-name

import collections
from typing import Dict, List, Tuple, Union
import warnings
import numpy as np
from fidescls.cls import models as _cls_models

AGGREGATION_METHODS = {"mean": np.nanmean, "max": np.nanmax, "median": np.nanmedian}

CLS_OUTPUT_TYPE = Union[
    List[_cls_models.MethodClassification],
    Dict[str, List[_cls_models.MethodClassification]],
]
AGG_CLS_OUTPUT_TYPE = Union[
    List[_cls_models.AggregatedMethodClassification],
    Dict[str, List[_cls_models.AggregatedMethodClassification]],
]


def validate_aggregation_method(method: str) -> None:
    """
    Check to see if the specified aggregation method is supported
    Args:
        method: the name of the aggregation method

    Raises:
        ValueError: if the provided aggregation method is not supported
    """
    if method not in AGGREGATION_METHODS.keys():
        raise ValueError(
            f"{method} is not a valid aggregation method!"
            f'Must be one of [{", ".join(AGGREGATION_METHODS.keys())}]'
        )


def aggregate_scores(
    cls_outputs: List[_cls_models.MethodClassification],
    aggregation_method: str,
    infer_not_pii: bool = True,
    paradigm: str = "content",
) -> List[_cls_models.AggregatedMethodLabel]:
    """

    Args:
        cls_outputs:
        aggregation_method:
        infer_not_pii: infer the presence of pii as a percentage of not pii labels to number of samples
        paradigm: classification paradigm from which the label was derived

    Returns:
        a list of the aggregated scores grouped by classification label
    """
    validate_aggregation_method(aggregation_method)

    grouped_results = collections.defaultdict(list)
    not_pii_count = 0

    for output in cls_outputs:
        if not output.labels:
            not_pii_count += 1
            continue
        for label in output.labels:
            grouped_results[label.label].append(label.score)

    aggregated_scores = []
    for label, scores in grouped_results.items():
        aggregated_scores.append(
            _cls_models.AggregatedMethodLabel(
                label=label,
                score=AGGREGATION_METHODS[aggregation_method](scores),  # type: ignore
                classification_paradigm=paradigm,
            )
        )
    if infer_not_pii and not_pii_count > 0:
        aggregated_scores.append(
            _cls_models.AggregatedMethodLabel(
                label=None,
                score=np.clip(
                    np.round(not_pii_count / len(cls_outputs), decimals=3), 0, 1
                ),
                classification_paradigm=paradigm,
            )
        )
    return aggregated_scores


def aggregate_classification(
    cls_output: CLS_OUTPUT_TYPE,
    aggregation_method: str,
    infer_not_pii: bool = True,
    paradigm: str = "content",
) -> AGG_CLS_OUTPUT_TYPE:
    """
    Aggregate classification output using the specified aggregation method
    Args:
        cls_output: classification output
        aggregation_method: aggregation method name
        infer_not_pii: flag to infer not pii class label score if classification
        method does not have such a class label
        paradigm: classification paradigm from which the label was derived

    Returns:

    """

    validate_aggregation_method(aggregation_method)

    aggregated_output: Union[
        List[_cls_models.AggregatedMethodClassification],
        Dict[str, List[_cls_models.AggregatedMethodClassification]],
    ]

    if isinstance(cls_output, dict):
        aggregated_output = {}
        for key, cls_results in cls_output.items():
            aggregated_output[key] = [
                _cls_models.AggregatedMethodClassification(
                    input=[i.input for i in cls_results],
                    labels=aggregate_scores(
                        cls_results,
                        aggregation_method=aggregation_method,
                        infer_not_pii=infer_not_pii,
                        paradigm=paradigm,
                    ),  # need to unpack/extend a list of lists i forget the function
                    aggregation_method=aggregation_method,
                )
            ]
    else:
        aggregated_output = []
        aggregated_output.append(
            _cls_models.AggregatedMethodClassification(
                input=[i.input for i in cls_output],
                labels=aggregate_scores(
                    cls_output,
                    aggregation_method=aggregation_method,
                    infer_not_pii=infer_not_pii,
                    paradigm=paradigm,
                ),  # need to unpack/extend a list of lists i forget the function
                aggregation_method=aggregation_method,
            )
        )
    return aggregated_output


def validate_system_aggregation_weights(
    context_weight: float = 0.55,
    content_weight: float = 0.45,
    prefer_context: bool = True,
    **kwargs: Dict,
) -> Tuple[float, float]:
    """
    Validate and/or set the system aggregation weights for
    context and content classification aggregation
    Args:
        context_weight: classification score multiplication factor
        content_weight: classification score multiplication factor
        prefer_context: boolean to prefer the weight associated with
        context if there's a conflict

    Returns:
        validated set of weights (context, content)

    """
    min_weight = 0.0
    max_weight = 1.0
    if (
        not min_weight < context_weight < max_weight
        or not min_weight < content_weight < max_weight
    ):
        raise ValueError(
            f"System aggregation weights must be [0.0, 1.0], {context_weight} and {content_weight} provided!"
        )
    if context_weight + content_weight != 1.0:
        warnings.warn(
            f"System aggregation weights ({context_weight} and {content_weight}) do not sum to 1.0."
            f'Using preferred weight: {"context" if prefer_context else "content"}'
        )
        if prefer_context:
            precision = str(context_weight)[::-1].find(".")
            content_weight = np.round(1.0 - context_weight, decimals=precision)
        else:
            precision = str(content_weight)[::-1].find(".")
            context_weight = np.round(1.0 - content_weight, decimals=precision)
    return context_weight, content_weight


def adjust_scores_by_weight(cls_output: List, weight: float) -> List:
    """
    Adjust classification scores by the specified weight
    Args:
        cls_output: the classification output to be adjusted
        weight: the weight by which to adjust the classification score

    Returns:
        The classification output with a new weighted score attribute
    """
    for classification in cls_output:
        for label in classification.labels:
            if label.score:
                label.aggregated_score = np.round(label.score * weight, decimals=4)
            else:
                label.aggregated_score = 0.0
    return cls_output


def merge_classifications(
    context_classification: List,
    content_classification: List,
    aggregation_method: str,
    aggregation_params: Dict,
    sort_attribute: str = "aggregated_score",
    desc: bool = True,
) -> _cls_models.AggregatedSystemClassification:
    """
    Merge content and context classification results together in a system aggregation data model
    Args:
        context_classification: results from a context classification
        content_classification: results from a content classification
        aggregation_method: name of the method by which to aggregate
        aggregation_params: parameters to be passed into the aggregation method
        sort_attribute:  data model attribute by which to sort classification labels
        desc: flag to set ordering direction

    Returns:

    """
    merged_classifications = _cls_models.AggregatedSystemClassification()
    merged_classifications.aggregation_method = aggregation_method
    merged_classifications.aggregation_params = aggregation_params

    context_inputs = []
    content_inputs = []
    for classification in context_classification:
        context_inputs.extend(
            classification.input
            if isinstance(classification.input, list)
            else [classification.input]
        )
        merged_classifications.aggregated_labels.extend(
            [
                _cls_models.AggregatedMethodLabel(
                    label=i.label,
                    score=i.score,
                    aggregated_score=i.aggregated_score,
                    classification_paradigm="context",
                )
                for i in classification.labels
            ]
        )
    merged_classifications.context_input = context_inputs
    for classification in content_classification:
        content_inputs.extend(
            classification.input
            if isinstance(classification.input, list)
            else [classification.input]
        )
        merged_classifications.aggregated_labels.extend(classification.labels)
    merged_classifications.content_input = content_inputs
    # score the labels by their weighted scores
    merged_classifications.aggregated_labels.sort(
        key=lambda x: (getattr(x, sort_attribute) is None, getattr(x, sort_attribute)),
        reverse=desc,
    )
    # perform de-duplication after sorting and keep the first occurrence of a label
    return merged_classifications


def deduplicate_aggregated_labels(
    labels: List[_cls_models.AggregatedMethodLabel],
) -> List[_cls_models.AggregatedMethodLabel]:
    """
    Remove duplicated labels in an aggregated set of classification labels; retaining
    the first occurrence
    Args:
        labels: list of labels to deduplicate

    Returns:

    """
    unique_labels: List = []
    for label in labels:
        if label.label not in {i.label for i in unique_labels}:
            unique_labels.append(label)
    return unique_labels


def aggregate_system(
    context_classification: List,
    content_classification: List,
    aggregation_method: str = "weighted",
    aggregation_params: Dict = {},
) -> _cls_models.AggregatedSystemClassification:
    """
    Aggregate context and content classification results
    Args:
        context_classification: results from a context classification
        content_classification: results from a content classification
        aggregation_method: name of the method by which to aggregate [default: 'weighted']
        aggregation_params: parameters to be passed into the aggregation method

    Returns:

    """
    # weighted aggregation is the only implementation, currently
    # when more methods get added, this will have to be abstracted
    if aggregation_method != "weighted":
        raise ValueError(f"Aggregation method, {aggregation_method}, not supported!")

    # get the aggregation weights
    context_weight, content_weight = validate_system_aggregation_weights(
        **aggregation_params
    )

    # adjust scores by weight factor
    adjust_scores_by_weight(context_classification, context_weight)
    adjust_scores_by_weight(content_classification, content_weight)

    aggregated_classifications = merge_classifications(
        context_classification,
        content_classification,
        aggregation_method,
        aggregation_params,
    )

    # return top n labels removing duplicated labels
    aggregated_classifications.aggregated_labels = deduplicate_aggregated_labels(
        aggregated_classifications.aggregated_labels
    )[: aggregation_params.get("top_n", 1)]

    return aggregated_classifications



"""
Configuration settings for the api
"""
from pathlib import Path

PORT = 8765
RESOURCES = ["text"]
CLASSIFY_ENDPOINT = "classify"
LOGGER_NAME = "fidescls_api"
LOG_FILENAME = Path(__file__).parent / "logs/api.log"

PII_DEFAULT_THRESHOLD = 0.3



"""
Utility functionality around supporting database interfaces and data ingestion
"""

from collections import defaultdict
import logging
from typing import Dict, Iterable, List
import click
from tqdm import tqdm
import pandas as pd
import sqlalchemy as sql

SCHEMA_EXCLUSION = {
    "postgresql": ["information_schema"],
    "mysql": ["mysql", "performance_schema", "sys", "information_schema"],
    "mssql": ["INFORMATION_SCHEMA", "guest", "sys"],
}


def get_engine(
    connection_string: str, test_connection: bool = True, verbose: bool = False
) -> sql.engine.Engine:
    """
    Get a sqlalchemy engine using a given connection string

    Args:
        connection_string: sqlalchemy connection string (https://docs.sqlalchemy.org/en/latest/core/engines.html)
        test_connection: attempt to pull from the database to test connection
        verbose: flag to enable status messages

    Returns:
        sqlalchemy database engine
    """
    logging.debug("Getting engine for database connection.")
    engine = sql.create_engine(connection_string)
    if test_connection:
        try:
            if verbose:
                click.secho("Testing database connection...", fg="white")
            engine.connect()
        except sql.exc.OperationalError as err:
            logging.exception(f"Could not connect to database! {err}")
            raise err
        if verbose:
            click.secho("Connection established!", fg="white")
    return engine


def get_database_type(
    engine: sql.engine.Engine,
    check_valid: bool = True,
    valid_database_types: Iterable = tuple(SCHEMA_EXCLUSION.keys()),
) -> str:
    """
    Get the name of the type of database given sqlalchemy engine
    Args:
        engine: The name of the type of database being connected to
        check_valid: Check to see if this database type is supported by fidescls
        valid_database_types: iterable

    Returns:
        name of the type of database the engine is pointing to
    """
    database_type = engine.dialect.name
    logging.debug(f"Database type: {database_type}")
    if check_valid:
        if database_type not in valid_database_types:
            raise ValueError(f"{database_type} is not currently supported!")
    return database_type


def include_schema(
    schema: str,
    database_type: str,
    exclusion_map: Dict[str, list[str]] = SCHEMA_EXCLUSION,
) -> bool:
    """
    Check to see if a schema should be included based on database type and schema name
    Args:
        schema: The name of the schema being checked
        database_type: the name of the database type
        exclusion_map: a dictionary mapping between database name and an exclusion list

    Returns:
        True if the schema should be included, False otherwise
    """
    if database_type == "mssql":
        return schema not in exclusion_map[database_type] and not schema.startswith(
            "db_"
        )
    return schema not in exclusion_map[database_type]


def get_table_key(schema: str, table: str, database_type: str) -> str:
    """
    Construct the table name key based on database type
    Args:
        schema: name of the schema
        table: name of the table
        database_type: name of the database type

    Returns:
        the appropriately formatted tabel name
    """
    return f"{schema}.{table}" if database_type == "postgresql" else table


def get_db_table_column_names(
    engine: sql.engine.Engine, hide_progress: bool = False
) -> Dict[str, Dict[str, List[str]]]:
    """
    Extract the table and column names from a database given a sqlalchemy engine
    Args:
        engine: sqlalchemy engine
        hide_progress: flag to disable progress bars and status messages

    Returns:
        Nested dictionary containing a database's tables and column names
    """
    logging.debug("Getting database name metadata...")
    inspector = sql.inspect(engine)
    database_type = get_database_type(engine)

    if not hide_progress:
        click.secho("Scanning database...", fg="white")

    db_tables: Dict[str, Dict[str, List]] = {}
    for schema in tqdm(
        inspector.get_schema_names(), disable=hide_progress, position=0, desc="Schema"
    ):
        if include_schema(schema, database_type):
            logging.debug(f"Extracting Schema: {schema}")
            db_tables[schema] = {}
            for table in tqdm(
                inspector.get_table_names(schema=schema),
                disable=hide_progress,
                position=1,
                desc="Tables",
                leave=False,
            ):
                # postgres table key name is handled differently for unknown reason at this time
                # table_key = f'{schema}.{table}' if database_type == 'postgresql' else table
                db_tables[schema][table] = [
                    column["name"]
                    for column in inspector.get_columns(table, schema=schema)
                ]
        else:
            logging.debug(f"Skipping schema: {schema}")
    return db_tables


def scan_database_metadata(
    connection_string: str, hide_progress: bool = False
) -> Dict[str, Dict[str, List[str]]]:
    """
    Scan a database for schema, table, and column names

    Args:
        connection_string: The sqlalchemy compatible connection string to the database
        hide_progress: flag to hide the progress bars

    Returns:
        The database's schema, tabel, and column names.
        Format: {schema_name: {table_name: [column names, ]}}
    """
    engine = get_engine(connection_string)
    metadata = get_db_table_column_names(engine, hide_progress=hide_progress)
    return metadata


def metadata_to_dataframe(
    database_metadata: Dict[str, Dict[str, List[str]]]
) -> pd.DataFrame:
    """
    Convert a metadata nested dictionary to a dense pandas dataframe to ease processing
    Args:
        database_metadata: nested dictionary containing database metadata (output of scan_database_metadata)

    Returns:
        dense dataframe containing a database's metadata
    """
    temp_list = []
    for schema, tables in database_metadata.items():
        for table, columns in tables.items():
            temp_list.append({"schema": schema, "table": table, "columns": columns})
    return pd.DataFrame.from_dict(temp_list)


def get_table_samples(
    engine: sql.engine.Engine, schema_name: str, table_name: str, num_samples: int = 10
) -> Dict[str, List[str]]:
    """
    Get a number of content samples from a table
    Args:
        engine: sqlalchemy engine with connection to the database
        schema_name: the name of the schema for the table
        table_name: the name of the table from which to query
        num_samples: the number of rows to pull from the table

    Returns:

    """
    with engine.begin() as connection:
        # TODO: figure out how to make this a random sample - abstracted for database type
        raw_query = sql.select("*", sql.table(table_name, schema=schema_name)).limit(  # type: ignore[arg-type, call-arg]
            num_samples
        )
        query_result = connection.execute(raw_query)
        query_columns = list(query_result.keys())
        query_samples = query_result.all()

    # reformat into an easier way to load into a pandas dataframe later
    # {column_name: [sample1, sample2, ..., sampleN]}
    table_samples: Dict[str, List[str]] = defaultdict(list)
    for i, column in enumerate(query_columns):
        for result in query_samples:
            table_samples[column].append(str(result[i]))
    return dict(table_samples)



"""
CLI command wrappers for database scanning and classification functionality
"""
import pprint
import json
import click

from fastapi.encoders import jsonable_encoder

from fidescls.cli import options as _opt


@click.group(name="db")
def db() -> None:
    """
    A CLI command group focused on database scanning for classification
    """
    # pylint: disable=invalid-name
    ...


@db.command(name="classify")
@click.pass_context
@_opt.num_samples
@_opt.use_context
@_opt.use_content
@_opt.use_aggregation
@_opt.aggregation_method
@click.argument("connection_string", type=str)
@click.option(
    "-f",
    "--filename",
    "file_name",
    type=str,
    nargs=1,
    default="",
    help="filename to save database classification output json",
)
@click.option(
    "-q",
    "--quiet",
    "quiet",
    type=bool,
    default=False,
    help="Disable progress bars and feedback",
)
@click.option(
    "-e",
    "--exclude-none",
    "exclude_none",
    type=bool,
    is_flag=True,
    default=False,
    help="Exclude null objects from classification output",
)
def classify(
    ctx: click.Context,
    connection_string: str,
    file_name: str,
    use_context: bool,
    use_content: bool,
    use_aggregation: bool,
    aggregation_method: str,
    num_samples: int,
    quiet: bool,
    exclude_none: bool,
) -> None:
    """
    Classify PII in each column of each table of a database
    """
    from fidescls.cls import scan as _scan

    db_cls_output = _scan.label_database(
        connection_string,
        classify_context=use_context,
        classify_content=use_content,
        aggregate_output=use_aggregation,
        aggregation_method=aggregation_method,
        num_samples=num_samples,
        hide_progress=quiet,
    )
    if file_name:
        with open(file_name, "w", encoding="utf-8") as output_file:
            json.dump(
                jsonable_encoder(db_cls_output, exclude_none=exclude_none), output_file
            )
        click.secho(f"Classification output saved to: {file_name}")
    else:
        click.secho(
            pprint.pformat(jsonable_encoder(db_cls_output, exclude_none=exclude_none)),
            fg="white",
        )
